/*
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//
// @author Marco Magistrali
// @version $Id: SyncMLDMBuilder.cpp,v 1.0 2004/04/27 11:21:00 magi Exp $
//

#include "spdm/win32/SyncMLDMBuilder.h"
#include "spdm/common/DMMessageTemplate.h" 
#include "spdm/common/md5.h"
#include "spdm/common/base64.h"
#include "spds/common/Utils.h"

USING_PTYPES

/*
 * Constructors
 */

SyncMLDMBuilder::SyncMLDMBuilder() {
   return;
}

wchar_t* SyncMLDMBuilder::createInitMessage(SyncMLDMConfig* syncMLDMConfig, 
                                            DevInfo* devInfo, wchar_t* dmSessionId, 
                                            int *cmdId, int *msgId,
                                            BOOL isClientNotified,
                                            wchar_t* authStatusMsg) {

    wchar_t codeNotification [10];
    if (isClientNotified == TRUE)
        wsprintf(codeNotification, TEXT("1200")); // 
    else
        wsprintf(codeNotification, TEXT("1201"));
    
    wchar_t* md5Data = NULL;
    int firstCmd = ++(*cmdId); 
    string devInfoMsg = string();
    
    createDevInfoMessage(devInfoMsg, devInfo, ++(*cmdId));
    
/* ---------------- for auth-clear  ------------------ */
/*
    int size = wcslen(DMxmlInit1Clear)        
         + 2048             
         ;

    wchar_t* buf = new wchar_t[size];
    *buf = 0;        

    wsprintf(buf                                            ,
             DMxmlInit1Clear                                ,
             dmSessionId                                    ,
             ++(*msgId)                                     ,
             syncMLDMConfig->syncMLDMAcc->getAddr()         ,
             devInfo->getDevId()                            ,
             syncMLDMConfig->syncMLDMAcc->getUsername()     ,
             syncMLDMConfig->syncMLDMAcc->getClientPW()     , 
             authStatusMsg                                  ,
             firstCmd                                       ,
             codeNotification                               
             );
    wcscat(buf, (wchar_t*)devInfoMsg);
    wcscat(buf, DMxmlInit2);
    return buf;
*/
/* ---------------- for auth-Md5 ------------------ */
   
    md5Data = createMD5CredentialData (syncMLDMConfig);
    int size = wcslen(DMxmlInit1Md5)        
         + 2048             
         ;

    wchar_t* buf = new wchar_t[size];
    *buf = 0;        

    wsprintf(buf                                            ,
             DMxmlInit1Md5                                  ,
             dmSessionId                                    ,
             ++(*msgId)                                     ,
             syncMLDMConfig->syncMLDMAcc->getAddr()         ,
             devInfo->getDevId()                            ,
             md5Data                                        ,
             authStatusMsg                                  ,
             firstCmd                                       ,
             codeNotification                               
             );
    wcscat(buf, (wchar_t*)devInfoMsg);
    wcscat(buf, DMxmlInit2);       
    
    //delete md5Data;
    return buf;


  
}


wchar_t* SyncMLDMBuilder::createResponseStatusMessage(SyncMLDMConfig* syncMLDMConfig, DevInfo* devInfo, 
                                    wchar_t* DMSessionId, 
                                    int *cmdId, int *msgId, wchar_t* responseStatusMsg, wchar_t* authStatusMsg) {
    
    int size = wcslen(DMxmlCmdStatus1)        
         + 2048             
         ;

    wchar_t* buf = new wchar_t[size];
    *buf = 0;        

    wsprintf(buf                                            ,
             DMxmlCmdStatus1                                ,
             DMSessionId                                    ,
             ++(*msgId)                                     ,
             syncMLDMConfig->syncMLDMAcc->getAddr()         ,
             devInfo->getDevId()                            ,
              ++(*cmdId)                                     
             );
    wcscat(buf, authStatusMsg);
    wcscat(buf, responseStatusMsg);
    wcscat(buf, DMxmlCmdStatus2);

    return buf;
}



wchar_t* SyncMLDMBuilder::createReplaceStatusMessage(int msgRef, wchar_t *cmdRef, int *cmdId, string &targetRef, int data, BOOL onlyOneStatus) {
    
    if (length(targetRef) == 0)
        return TEXT("");

    int size = wcslen(DMxmlReplaceStatus)        
         + 2048              
         ;

    wchar_t* buf = new wchar_t[size];
    *buf = 0;        
     wsprintf(buf                 ,
             DMxmlReplaceStatus   ,
             ++(*cmdId)           ,
             msgRef               ,
             cmdRef               ,             
            (onlyOneStatus == TRUE) ? TEXT("") : (wchar_t*)targetRef   ,
             data             
             );
    
    return buf;
}



wchar_t* SyncMLDMBuilder::createSyncHdrStatusMessage(int *cmdId, int serverMsgId, SyncMLDMConfig* syncMLDMConfig, DevInfo* devInfo) {
    
    wchar_t* buf = new wchar_t [2048];
    *buf = 0;
    
    wsprintf(buf,
        DMxmlAuthStatus,
        ++(*cmdId),
        serverMsgId,
        TEXT("0"),        
        devInfo->getDevId(),
        syncMLDMConfig->syncMLDMAcc->getAddr()
        );
    return buf;
}


wchar_t* SyncMLDMBuilder::createExecStatusMessage(int msgRef, wchar_t *cmdRef, int *cmdId, string& targetRef, int data, BOOL onlyOneStatus) {
    //wchar_t* createExecStatusMessage(int msgRef, wchar_t *cmdRef,
    //                                int *cmdId,  wchar_t* locUriData);

    if (length(targetRef) == 0)
        return TEXT("");


    int size = wcslen(DMxmlExecStatus)        
         + 2048             
         ;

    wchar_t* buf = new wchar_t[size];
    *buf = 0;        
     wsprintf(buf                 ,
             DMxmlExecStatus   ,
             ++(*cmdId)           ,
             msgRef               ,
             cmdRef               ,
             (onlyOneStatus == TRUE) ? TEXT("") : (wchar_t*)targetRef   ,
             data
             );
    
    return buf;
}

wchar_t* SyncMLDMBuilder::createGetStatusMessage(int msgRef, wchar_t *cmdRef, int *cmdId, string &targetRef, int data, BOOL onlyOneStatus) {
    
    
    if (length(targetRef) == 0)
        return TEXT("");

    int size = wcslen(DMxmlGetStatus)        
         + 2048              
         ;

    wchar_t* buf = new wchar_t[size];
    *buf = 0;        
     wsprintf(buf                                                       ,
             DMxmlGetStatus                                             ,
             ++(*cmdId)                                                 ,
             msgRef                                                     ,
             cmdRef                                                     ,
             (onlyOneStatus == TRUE) ? TEXT("") : (wchar_t*)targetRef   ,
             data
             );
    return buf;

    
}

wchar_t* SyncMLDMBuilder::createResultsMessage(wchar_t* locUri, wchar_t* data, BOOL isNode) {
    
    int size = wcslen(DMxmlResults)        
         + 2048              
         ;
    wchar_t tmp [1024];
    wsprintf(tmp, DMxmlResultsEnd, data);
    wchar_t* buf = new wchar_t[size];
    *buf = 0;        
     wsprintf(buf                 ,
             DMxmlResults         ,
             locUri               
             );
    if (isNode == TRUE)
        wcscat(buf, DMxmlResultsIsNode);
    
    wcscat(buf, tmp);
    
    return buf;


}

void SyncMLDMBuilder::completeResults(string &results, int msgRef, wchar_t *cmdRef, int cmdId ) {
    
   
    DMxmlCompleteResults = TEXT("<Results>\n<CmdID>%i</CmdID>\n<MsgRef>%i</MsgRef>\n<CmdRef>%s</CmdRef>\n");
    wchar_t tmp [1024];
    wsprintf(tmp, DMxmlCompleteResults, cmdId, msgRef, cmdRef);
    ins (tmp, results, 0);
    results += TEXT("</Results>\n");


}

/*------------------------------------------------- private methods -------*/

void SyncMLDMBuilder::createDevInfoMessage(string &s, DevInfo *devInfo, int cmd) {
    
    wchar_t cmdId[10];
    wsprintf(cmdId, TEXT("%i"), cmd);

    s += TEXT("<Replace>\n<CmdID>");
    s += cmdId;
    s += TEXT("</CmdID>\n");
    //DevId
    s += TEXT("<Item>\n");
    s += TEXT(  "<Source><LocURI>./DevInfo/DevId</LocURI></Source>\n");
    s += TEXT(  "<Meta>\n");
    s += TEXT(      "<Format xmlns='syncml:metinf'>chr</Format>\n");
    s += TEXT(      "<Type xmlns='syncml:metinf'>text/plain</Type>\n");
    s += TEXT(  "</Meta>\n");
    s += TEXT(  "<Data>");
    s +=        (wchar_t*)devInfo->getDevId();
    s += TEXT(  "</Data>\n");
    s += TEXT("</Item>\n");

    //Man
    s += TEXT("<Item>\n");
    s += TEXT(  "<Source><LocURI>./DevInfo/Man</LocURI></Source>\n");
    s += TEXT(  "<Meta>\n");
    s += TEXT(      "<Format xmlns='syncml:metinf'>chr</Format>\n");
    s += TEXT(      "<Type xmlns='syncml:metinf'>text/plain</Type>\n");
    s += TEXT(  "</Meta>\n");
    s += TEXT(  "<Data>");
    s +=        (wchar_t*)devInfo->getMan();
    s += TEXT(  "</Data>\n");
    s += TEXT("</Item>\n");
    
    // Mod
    s += TEXT("<Item>\n");
    s += TEXT(  "<Source><LocURI>./DevInfo/Mod</LocURI></Source>\n");
    s += TEXT(  "<Meta>\n");
    s += TEXT(      "<Format xmlns='syncml:metinf'>chr</Format>\n");
    s += TEXT(      "<Type xmlns='syncml:metinf'>text/plain</Type>\n");
    s += TEXT(  "</Meta>\n");
    s += TEXT(  "<Data>");
    s +=        (wchar_t*)devInfo->getMod();
    s += TEXT(  "</Data>\n");
    s += TEXT("</Item>\n");
    
    //DmV
    s += TEXT("<Item>\n");
    s += TEXT(  "<Source><LocURI>./DevInfo/DmV</LocURI></Source>\n");
    s += TEXT(  "<Meta>\n");
    s += TEXT(      "<Format xmlns='syncml:metinf'>chr</Format>\n");
    s += TEXT(      "<Type xmlns='syncml:metinf'>text/plain</Type>\n");
    s += TEXT(  "</Meta>\n");
    s += TEXT(  "<Data>");
    s +=        (wchar_t*)devInfo->getDmv();
    s += TEXT(  "</Data>\n");
    s += TEXT("</Item>\n");
    
    //Lang
    s += TEXT("<Item>\n");
    s += TEXT(  "<Source><LocURI>./DevInfo/Lang</LocURI></Source>\n");
    s += TEXT(  "<Meta>\n");
    s += TEXT(      "<Format xmlns='syncml:metinf'>chr</Format>\n");
    s += TEXT(      "<Type xmlns='syncml:metinf'>text/plain</Type>\n");
    s += TEXT(  "</Meta>\n");
    s += TEXT(  "<Data>");
    s +=        (wchar_t*)devInfo->getLang();
    s += TEXT(  "</Data>\n");
    s += TEXT("</Item>\n");

    s += TEXT("</Replace>");

}


/*
* It implements algo for authentication with MD5 method.
* It computes digest token according with follow:
* Let H   : MD5 Function represents by calculateMD5 method
* Let B64 : Base64 encoding Function represents by encodeBase64 method
* Data: H (B64(H(username:password)):nonce)
*/

wchar_t*  SyncMLDMBuilder::createMD5CredentialData(SyncMLDMConfig *syncMLDMConfig) {        
    
    int lung = 0;

    char token [64];
    char nonce [64];    
    char digest [16];
    char base64 [64];
    char base64Nonce [64];
    
    memset(digest, 0, 16);
    memset(base64, 0, 64);
    memset(base64Nonce, 0, 64);
    memset(nonce, 0, 64);
    memset(token, 0, 64);
    
    wchar_t* md5Digest = NULL;
    wchar_t wnonce  [64 ];

    wchar_t wtoken [64];
    
    wmemset(wtoken, 0,64);

    wsprintf(wtoken, TEXT("%s:%s"), syncMLDMConfig->syncMLDMAcc->getUsername()
                                  , syncMLDMConfig->syncMLDMAcc->getClientPW()
                                  );

    lung = wcslen(wtoken);
    
    BOOL usedDefaultChar;
    
        
    WideCharToMultiByte (CP_ACP, 0, wtoken, lung, token, lung, "?", &usedDefaultChar);      
    token[lung] = 0;
        
    // H(username:password)
    calculateMD5(token, digest);
  
    // B64(H(username:password))    
    encodeBase64(digest, base64);
    memset(digest, 0, 16);   

    wsprintf(wnonce, TEXT("%s"), syncMLDMConfig->syncMLDMAcc->getClientNonce());    
    lung = wcslen(wnonce);
    

    WideCharToMultiByte (CP_ACP, 0, wnonce, lung, nonce, lung, "?", &usedDefaultChar);
    nonce[lung] = 0;

    // decode wnonce from stored base64 to bin
    decodeBase64(nonce, nonce);

    sprintf(base64Nonce, "%s:%s", base64, nonce);    
    base64Nonce[strlen(base64Nonce)] = 0;

    calculateMD5(base64Nonce, digest);
    memset(base64, 0, 64);

    encodeBase64(digest, base64); 
  
    // return value in wchar...    
    md5Digest = new wchar_t[33];
    MultiByteToWideChar (CP_ACP, 0, base64, 33, md5Digest, 33);
    md5Digest[33] = 0;
    
    return md5Digest;

}

/*
* It calculates MD5 digest from a token. It's used by createMD5CredentialData.
*/

// void SyncMLDMBuilder::calculateMD5(wchar_t* wtoken, wchar_t* wdigest) {
void SyncMLDMBuilder::calculateMD5(char* token, char* wdigest) {

    int lung = strlen(token);
    //algo for md5 digest

    md5_state_t state;
    md5_byte_t digest[16];
    // char hex_output[16*2 + 1];
    // int di;
    int i, k;
    char dig [16];
    memset(dig, 0, 16);
    md5_init(&state);
    md5_append(&state, (const md5_byte_t *)token, lung);
    md5_finish(&state, digest);
    
    //
    // To retrieve exadecimal data
    //

    //for (di = 0; di < 16; ++di)
    //    sprintf(hex_output + di * 2, "%02x", digest[di]);
    
    
    for (i = 0; i <= 15; i++) {
        k = i;
        sprintf(dig + i, "%c", digest[i]);
        if (k == 15)
            break;
    }
    strcpy(wdigest, dig);

    // end algo     
               
}

void SyncMLDMBuilder::encodeBase64(char *token, char *base64) {
        
    int lung = strlen(token);    
    char *dest = new char[4 * lung];    
    memset(dest,       0, 4 * lung);
    
    int rett = b64_encode(dest, token, lung);

    strcpy (base64, dest);
    delete dest;
     
}

void SyncMLDMBuilder::decodeBase64(char *base64, char *token) {
        
    int lung = strlen(base64);    
    char *dest = new char[4 * lung];    
    memset(dest,       0, 4 * lung);
    
   
    int rett = b64_decode(dest, base64);

    strcpy (token, dest);
    delete dest;
     
}


